package rhymes;

import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * This class defines a node of a suffix tree. The node holds a character, a set
 * of references to its children and a reference to its parent. Moreover, the
 * node contains information, whether it is starting point of a correct word.
 *
 * @author Daniel Oertwig
*/
public class SuffixNode {
    /**
     * The parent node of this node. Every node except the root node has a
     * parent.
    */
    private SuffixNode parent;
    /**
     * The character represented by this node inside the tree.
    */
    private char myChar;
    /**
     * Variable number of children.
    */
    private HashMap<Character, SuffixNode> children;
    /**
     * This boolean indicates, whether the node is starting point of a correct
     * word.
    */
    private boolean wordNode;
    
    
    /**
     * Constructor. Creates a node representing character c with parent par.
     * Furthermore, the constructor needs to know, whether the node defines a
     * correct word.
     *
     * @param c The character which is represented by this node.
     * @param isWordNode Boolean indicating, whether this node defines a correct
     * word.
     * @param par The parent node of this node. When constructing the root node,
     * this parameter can be null.
    */
    public SuffixNode(char c, boolean isWordNode, SuffixNode par) {
        this.parent = par;
        this.myChar = c;
        this.wordNode = isWordNode;
    }
    
    /**
     * Returns the character, which is represented by this node.
     *
     * @return The character represented by this node.
    */
    public char getChar() {
        char character = '#';
        if (this.parent != null) {
            character = this.myChar;
        }
        return character;
    }
    
    /**
     * Tests whether this node has a child, which represents the character c.
     *
     * @param c The character to search for.
     *
     * @return True if a child representing c is found, false otherwise.
    */
    public boolean hasChild(char c) {
        boolean childExists = false;
        if (this.children != null) {
            childExists = this.children.containsKey(c);
        }
        return childExists;
    }
    
    /**
     * Returns the child, which represents the character c.
     *
     * @param c The character to search for.
     *
     * @return The child representing the character, or null if no such child
     * exists.
    */
    public SuffixNode getChild(char c) {
        SuffixNode child = null;
        if (this.children != null) {
            child = this.children.get(c);
        }
        return child;
    }
    
    /**
     * Checks, whether this node is the root node. The root node is the node,
     * which has no parent.
     *
     * @return True if this is the root node, false otherwise.
    */
    public boolean isRootNode() {
        boolean iAmRoot = false;
        if (this.parent == null) {
            iAmRoot = true;
        }
        return iAmRoot;
    }
    
    /**
     * Returns the root node of the tree containing this node.
     *
     * @return The root node.
    */
    public SuffixNode getRootNode() {
        SuffixNode rootNode = this;
        while (rootNode.parent != null) {
            rootNode = rootNode.parent;
        }
        return rootNode;
    }
    
    /**
     * Indicates, whether this node defines a correct word.
     *
     * @return True if this node is starting point of a correct word, false
     * otherwise.
    */
    public boolean isValidWord() {
        return this.wordNode;
    }
    
    /**
     * Returns the suffix of this node. The suffix is the concatenation of all
     * characters from this node up to the root node.
     *
     * @return The suffix of this node.
    */
    public String getSuffix() {
        StringBuffer suffix = new StringBuffer();
        SuffixNode traveller = this;
        while (traveller.parent != null) {
            suffix.append(traveller.myChar);
            traveller = traveller.parent;
        }
        return suffix.toString();
    }
    
    /**
     * Inserts a word relativ to this node. The resulting word will be the
     * concatenation of the String w and the suffix of this node.
     *
     * @param w The word to insert relativ to this node.
     *
     * @return True if new nodes have been created, false otherwise.
     *
     * @throws IllegalArgumentException Thrown in case Argument w is null.
    */
    public boolean insertWordRelatively(String w)
        throws IllegalArgumentException {
        boolean createdNodes = false;
        if (w == null) {
            throw new IllegalArgumentException("Argument 'String w' == null");
        }
        else if (w.length() != 0) {
            SuffixNode traveller = this;
            char actualChar;
            Character actualCharacter;
            for (int iterator = w.length() - 1; iterator >= 0; iterator--) {
                if (traveller.children == null) {
                    traveller.children = new HashMap<Character, SuffixNode>();
                }
                actualChar = w.charAt(iterator);
                actualCharacter = new Character(actualChar);
                if (!traveller.children.containsKey(actualCharacter)) {
                    traveller.children.put(actualCharacter,
                                           new SuffixNode(actualChar,
                                                          false,
                                                          traveller));
                    createdNodes = true;
                }
                traveller = traveller.children.get(actualCharacter);
            }
            if (!traveller.wordNode) {
                createdNodes = true;
                traveller.wordNode = true;
            }
        }
        return createdNodes;
    }
    
    /**
     * Searches for the node, which is relative to this node represented by the
     * string suffix.
     *
     * @param suffix The suffix, which represents the node we are looking for.
     *
     * @return The node which is relative to this node represented by the string
     * suffix, or null, if such a node does not exist.
     *
     * @throws IllegalArgumentException Thrown in case suffix is null
    */
    public SuffixNode getRelativeNode(String suffix)
        throws IllegalArgumentException {
        SuffixNode relativeNode = null;
        if (suffix == null) {
            throw new IllegalArgumentException("Argument "
                                               + "'String suffix' == null");
        }
        else {
            relativeNode = this;
            for (int iterator = suffix.length() - 1;
                (iterator >= 0) && (relativeNode != null);
                iterator--) {
                if (relativeNode.children != null) {
                    relativeNode = 
                        relativeNode.children.get
                        (new Character(suffix.charAt(iterator)));
                }
                else {
                    relativeNode = null;
                }
            }
        }
        return relativeNode;
    }
    
    /**
     * Searches for all successors of this node. So, all words which end to the
     * suffix of this node are going be returned as a list.
     *
     * @return List of all words which end to the suffix of this node.
    */
    public List<String> getAllSuccessors() {
        List<String> successors = new ArrayList<String>();
        String mySuffix = this.getSuffix();
        if (this.wordNode) {
            successors.add(mySuffix);
        }
        if (this.children != null) {
            Iterator iterator = this.children.values().iterator();
            while (iterator.hasNext()) {
                SuffixNode child = (SuffixNode) iterator.next();
                successors.addAll(child.getAllSuccessors());
            }   
        }
        return successors;
    }
}

