package rhymes;

import java.util.List;
import java.util.ArrayList;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

/**
 * This class defines a suffix tree. Words can be stored inside the tree. Words
 * which have equal endings go into the same subtree. Thus this class can be
 * used e.g. to detect rhymes.
 *
 * @author Daniel Oertwig
*/
public class SuffixTree {
    private static final int TO_LOWER_CONSTANT = 32;
    /**
     * The root node of the tree. This node contains a special character # which
     * is only used as a stopper element.
    */
    private SuffixNode wordStopper;
    /**
     * This private member indicates, how many (real) words are contained inside
     * the tree.
    */
    private long knownWords;
    
    
    /**
     * Constructor. Generates the root/stopper node an sets knownWords to 0.
    */
    public SuffixTree() {
        this.wordStopper = new SuffixNode('#', false, null);
        this.knownWords = 0;
    }
    
    /**
     * This method adds the word defined by String s to the tree. The word in s
     * is assumed to be a correct word. The string is searched for illegal
     * characters.
     *
     * @param s The word, which will be added. Must contain only legal
     * characters.
     *
     * @throws IllegalArgumentException This exception is thrown in one of the
     * following cases: Either s is a null pointer or s contains illegal
     * characters.
    */
    public void addWord(String s) throws IllegalArgumentException {
        if (s == null) {
            throw new IllegalArgumentException("Argument 'String s' == null");
        }
        else {
            boolean correctString = true;
            for (int iterator = s.length() - 1; iterator >= 0; iterator--) {
                char actualChar = s.charAt(iterator);
                if (((actualChar < 'a') || (actualChar > 'z'))
                    && (actualChar != 'ä')
                    && (actualChar != 'ö')
                    && (actualChar != 'ü')
                    && (actualChar != 'ß')) {
                    correctString = false;
                    break;
                }
            }
            if (correctString) {
                if (this.wordStopper.insertWordRelatively(s)) {
                    this.knownWords++;
                }
            }
            else {
                throw new IllegalArgumentException
                ("Argument 'String' hast wrong format");
            }
        }
    }
    
    /**
     * Returns the saved number of known words contained by the tree.
     *
     * @return Number of known words
    */
    public long getNumberOfKnownWords() {
        return this.knownWords;
    }
    
    /**
     * Searches for the node with suffix suffix inside the tree. In case such
     * node exists, the node is returned, otherwise null is returned.
     *
     * @param suffix The suffix to search for.
     *
     * @return The node defined by suffix, or null, if non existent.
     *
     * @throws IllegalArgumentException Thrown when suffix is a null pointer.
    */
    public SuffixNode getNode(String suffix) throws IllegalArgumentException {
        return this.wordStopper.getRelativeNode(suffix);
    }
    
    /**
     * Finds all words which end by a specific suffix.
     *
     * @param s The ending of the searched words.
     *
     * @return A list of words matching the search. The list can be empty. The
     * method will NOT return a null pointer.
    */
    public List<String> getWordsWithSuffix(String s) {
        List<String> words = new ArrayList<String>();
        try {
            SuffixNode suffixNode = this.getNode(s);
            if (suffixNode != null) {
                words.addAll(suffixNode.getAllSuccessors());
            }
        } catch (IllegalArgumentException illegalArgument) {
            illegalArgument.printStackTrace();
            System.err.println("[ERROR] "
                               + "Invalid argument for method "
                               + "getWordsWithSuffix!");
        }
        return words;
    }
    
    /**
     * Private Methode to clean a text from illegal characters. All illegal
     * characters are converted into spaces.
     *
     * @param text The input text.
     *
     * @return The cleaned text.
     *
     * @throws IllegalArgumentException Thrown in case text is null.
    */
    private String cleanText(String text) throws IllegalArgumentException {
        String clean = "";
        if (text == null) {
            throw new IllegalArgumentException("Argument "
                                               + "'String text' == null");
        }
        else {
            int textLen = text.length();
            StringBuffer buffer = new StringBuffer(textLen);
            for (int iterator = 0; iterator < textLen; iterator++) {
                char actualChar = text.charAt(iterator);
                if (((actualChar >= 'a') && (actualChar <= 'z'))
                    || (actualChar == '\u00E4')
                    || (actualChar == '\u00F6')
                    || (actualChar == '\u00FC')
                    || (actualChar == '\u00DF')) {
                    buffer.append(actualChar);
                }
                else if (((actualChar >= 'A') && (actualChar <= 'Z'))
                    || (actualChar == '\u00C4')
                    || (actualChar == '\u00D6')
                    || (actualChar == '\u00DC')) {
                    actualChar = (char) (actualChar + this.TO_LOWER_CONSTANT);
                    buffer.append(actualChar);
                }
                else {
                    buffer.append(' ');
                }
            }
            clean = buffer.toString();
        }
        return clean;
    }
    
    /**
     * Extract words from a text and add them into the tree. In this context, a
     * word is a substring of text, which does not contain a space, the
     * character before and after the substring must be a space or out of bounds
     * of the input string.
     *
     * @param text The input text containing the words.
     *
     * @throws IllegalArgumentException Thrown in case text is null.
    */
    public void extractWords(String text) throws IllegalArgumentException {
        if (text == null) {
            throw new IllegalArgumentException("Argument "
                                               + "'String text' == null");
        }
        else {
            String clean = this.cleanText(text);
            int textLen = clean.length();
            int iterator = 0;
            while (iterator < textLen) {
                int next = clean.indexOf(' ', iterator);
                if ((next - iterator) > 0) {
                    this.addWord(clean.substring(iterator, next));
                }
                else if ((iterator - next) > 0) {
                    this.addWord(clean.substring(iterator));
                    next = textLen;
                }
                iterator = next + 1;
            }
        }
    }
    
    /**
     * Extract the words found in a file.
     *
     * @see rhymes.SuffixTree#extractWords
     *
     * @param fileName The file name of the file containing the words.
     *
     * @throws IllegalArgumentException Thrown in case fileName is null.
    */
    public void extractWordsFromFile(String fileName)
        throws IllegalArgumentException {
        if (fileName == null) {
            throw new IllegalArgumentException("Argument "
                                               + "'String fileName' == null");
        }
        else {
            FileInputStream inFile;
            try {
                inFile = new FileInputStream(fileName);
            } catch (FileNotFoundException fnfe) {
                fnfe.printStackTrace();
                System.err.println("[ERROR] File " + fileName + " not found!");
                return;
            }
            InputStreamReader inReader = new InputStreamReader(inFile);
            BufferedReader bufReader = new BufferedReader(inReader);
            StringBuffer fileContent = new StringBuffer();
            String line;
            try {
                while ((line = bufReader.readLine()) != null) {
                    fileContent.append(line + ' ');
                }
            } catch (IOException ioe) {
                ioe.printStackTrace();
                System.err.println("[ERROR] Reading from "
                                   + fileName + " failed!");
            }
            String content = fileContent.toString();
            this.extractWords(content);
        }
    }
}

